from .interfaces import Model, TranslationModel, BilinearModel

from .translation import TransEModel
from .translation import TransHModel
from .translation import TransRModel
from .translation import TransDModel
from .translation import TorusEModel

from .bilinear import RESCALModel
from .bilinear import DistMultModel
from .bilinear import HolEModel
from .bilinear import ComplExModel
from .bilinear import AnalogyModel

from .deep import ConvKBModel
